/*
Author: Cao Thanh Tung
Date: 21/01/2010

File Name: pbaCVDHost.cu

===============================================================================

Copyright (c) 2010, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/pba.html

If you use PBA and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#include <stdio.h>
#include <device_functions.h>

#include "pbaCVD.h"

#define MAX_ITERATION   1000
#define THRESHOLD       0       // CVD Threshold. 

// Parameters for CUDA kernel executions
#define BLOCKX		16
#define BLOCKY		16
#define BLOCKSIZE	64
#define TILE_DIM	32
#define BLOCK_ROWS	8
#define BAND		256			// For simplicity, just assume we never need to work with a smaller texture.

/****** Global Variables *******/
short2 **pbaTextures;       // Two textures used to compute 2D Voronoi Diagram
float *pbaDensity; 
float *pbaPrefixWeight, *pbaTotalWeight; 
float *pbaPrefixX, *pbaPrefixY, *pbaTotalX, *pbaTotalY; 

int *pbaFlag_d, pbaFlag_h; 
short2 *pbaVoronoi, *pbaTemp; 

int pbaBuffer;              // Current buffer
int pbaMemSize;             // Size (in bytes) of a texture
int pbaTexSize;             // Texture size (squared texture)

texture<short2> pbaTexColor; 
texture<short2> pbaTexLinks; 
texture<short> pbaTexShort; 
texture<float> pbaTexFloatX, pbaTexFloatY, pbaTexFloatW;

/********* Kernels ********/
#include "pbaCVDKernel.h"

///////////////////////////////////////////////////////////////////////////
//
// Initialize necessary memory for 2D Voronoi Diagram computation
// - textureSize: The size of the Discrete Voronoi Diagram (width = height)
//
///////////////////////////////////////////////////////////////////////////
void pbaCVDInitialization(int textureSize)
{
    pbaTexSize = textureSize; 
    pbaMemSize = pbaTexSize * pbaTexSize * sizeof(short2); 

    pbaTextures = (short2 **) malloc(2 * sizeof(short2 *)); 

    // Allocate 2 textures
    cudaMalloc((void **) &pbaTextures[0], pbaMemSize); 
    cudaMalloc((void **) &pbaTextures[1], pbaMemSize); 

    // Memory to compute CVD
    cudaMalloc((void **) &pbaDensity, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaTotalWeight, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaTotalX, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaTotalY, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaPrefixX, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaPrefixY, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaPrefixWeight, pbaTexSize * pbaTexSize * sizeof(float)); 
    cudaMalloc((void **) &pbaFlag_d, sizeof(int)); 
}

///////////////////////////////////////////////////////////////////////////
//
// Deallocate all allocated memory
//
///////////////////////////////////////////////////////////////////////////
void pbaCVDDeinitialization()
{
    cudaFree(pbaTextures[0]); 
    cudaFree(pbaTextures[1]); 
    cudaFree(pbaDensity); 
    cudaFree(pbaPrefixWeight); 
    cudaFree(pbaTotalWeight); 
    cudaFree(pbaPrefixX); 
    cudaFree(pbaPrefixY); 
    cudaFree(pbaTotalX); 
    cudaFree(pbaTotalY); 

    cudaFree(pbaFlag_d); 

    free(pbaTextures); 
}

// Copy input to GPU 
void pba2DInitializeInput(short *input, float *density)
{
    cudaMemcpy(pbaTextures[0], input, pbaMemSize, cudaMemcpyHostToDevice); 
    cudaMemcpy(pbaDensity, density, pbaTexSize * pbaTexSize * sizeof(float), cudaMemcpyHostToDevice); 

    pbaBuffer = 0; 
}

// In-place transpose a squared texture. 
// Block orders are modified to optimize memory access. 
// Point coordinates are also swapped. 
void pba2DTranspose(short2 *texture)
{
    dim3 block(TILE_DIM, BLOCK_ROWS); 
    dim3 grid(pbaTexSize / TILE_DIM, pbaTexSize / TILE_DIM); 

    cudaBindTexture(0, pbaTexColor, texture); 
    kernelTranspose<<< grid, block >>>(texture, pbaTexSize); 
    cudaUnbindTexture(pbaTexColor); 
}

// Phase 1 of PBA. m1 must divides texture size
void pba2DPhase1(int m1) 
{
    dim3 block = dim3(BLOCKSIZE);   
    dim3 grid = dim3(pbaTexSize / block.x, m1); 

    // Flood vertically in their own bands
    cudaBindTexture(0, pbaTexColor, pbaTextures[pbaBuffer]); 
    kernelFloodDown<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize, pbaTexSize / m1); 
    cudaUnbindTexture(pbaTexColor); 

    cudaBindTexture(0, pbaTexColor, pbaTextures[1 - pbaBuffer]); 
    kernelFloodUp<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize, pbaTexSize / m1); 

    // Passing information between bands
    grid = dim3(pbaTexSize / block.x, m1); 
    kernelPropagateInterband<<< grid, block >>>(pbaTextures[pbaBuffer], pbaTexSize, pbaTexSize / m1); 

    cudaBindTexture(0, pbaTexLinks, pbaTextures[pbaBuffer]); 
    kernelUpdateVertical<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize, m1, pbaTexSize / m1); 
    cudaUnbindTexture(pbaTexLinks); 
    cudaUnbindTexture(pbaTexColor); 

    pbaBuffer = 1 - pbaBuffer; 
}

// Phase 2 of PBA. m2 must divides texture size
void pba2DPhase2(int m2) 
{
    // Compute proximate points locally in each band
    dim3 block = dim3(BLOCKSIZE);   
    dim3 grid = dim3(pbaTexSize / block.x, m2); 
    cudaBindTexture(0, pbaTexColor, pbaTextures[pbaBuffer]); 
    kernelProximatePoints<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize, pbaTexSize / m2); 

    cudaBindTexture(0, pbaTexLinks, pbaTextures[1 - pbaBuffer]); 
    kernelCreateForwardPointers<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize, pbaTexSize / m2); 

    // Repeatly merging two bands into one
    for (int noBand = m2; noBand > 1; noBand /= 2) {
        grid = dim3(pbaTexSize / block.x, noBand / 2); 
        kernelMergeBands<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize, pbaTexSize / noBand); 
    }

    // Replace the forward link with the X coordinate of the seed to remove
    // the need of looking at the other texture. We need it for coloring.
    grid = dim3(pbaTexSize / block.x, pbaTexSize); 
    kernelDoubleToSingleList<<< grid, block >>>(pbaTextures[1 - pbaBuffer], pbaTexSize); 
    cudaUnbindTexture(pbaTexLinks); 
    cudaUnbindTexture(pbaTexColor); 
}

// Phase 3 of PBA. m3 must divides texture size
void pba2DPhase3(int m3) 
{
    dim3 block = dim3(BLOCKSIZE / m3, m3); 
    dim3 grid = dim3(pbaTexSize / block.x); 
    cudaBindTexture(0, pbaTexColor, pbaTextures[1 - pbaBuffer]); 
    kernelColor<<< grid, block >>>(pbaTextures[pbaBuffer], pbaTexSize); 
    cudaUnbindTexture(pbaTexColor); 
}

void pba2DCompute(int phase1Band, int phase2Band, int phase3Band)
{
    // Vertical sweep
    pba2DPhase1(phase1Band); 

    pba2DTranspose(pbaTextures[pbaBuffer]); 

    // Horizontal coloring
    pba2DPhase2(phase2Band); 

    // Color the rows. 
    pba2DPhase3(phase3Band); 

    pba2DTranspose(pbaTextures[pbaBuffer]); 

    pbaVoronoi = pbaTextures[pbaBuffer]; 
    pbaTemp = pbaTextures[1 - pbaBuffer]; 
}

void pbaCVDComputeWeightedPrefix() 
{
    dim3 block = dim3(BLOCKSIZE); 
    dim3 grid = dim3(pbaTexSize); 

    int ns = BLOCKSIZE * 3 * sizeof(float); 

    cudaBindTexture(0, pbaTexFloatW, pbaDensity); 
    kernelComputeWeightedPrefixX<<< grid, block, ns >>>(pbaPrefixWeight, 
        pbaPrefixX, pbaPrefixY, pbaTexSize); 
    cudaUnbindTexture(pbaTexFloatW); 
}

void pbaCVDComputeCentroid() 
{
	dim3 block(BLOCKSIZE);	
    dim3 grid(pbaTexSize);

    // Construct 1D Voronoi diagram
    int ns = pbaTexSize * sizeof(short); 

    cudaBindTexture(0, pbaTexColor, pbaVoronoi); 
    kernelVoronoi1D<<< grid, block, ns >>>((int *) pbaTemp, pbaTexSize); 

    // Sum along X axis
	cudaBindTexture(0, pbaTexFloatX, pbaPrefixX); 
    cudaBindTexture(0, pbaTexFloatY, pbaPrefixY); 
    cudaBindTexture(0, pbaTexFloatW, pbaPrefixWeight); 

	kernelTotal_X<<< grid, block >>>(pbaTotalX, pbaTotalY, pbaTotalWeight, pbaTexSize); 

	cudaUnbindTexture(pbaTexFloatX); 
	cudaUnbindTexture(pbaTexFloatY); 
	cudaUnbindTexture(pbaTexFloatW); 
    
    // Sum along Y axis. Up and down are done at the same time. 
	// Result of VC at (x, y) will be the sum over three pixels: 
	//	(x, y), (x, y-1) and (x, y+1)
    block = dim3(BLOCKSIZE); 
    grid = dim3(pbaTexSize / block.x, 2); 

    cudaBindTexture(0, pbaTexShort, pbaTemp); 

	kernelTotalWeight_Y<<< grid, block >>>(pbaTotalWeight, pbaTotalX, pbaTotalY, pbaTexSize); 

	cudaUnbindTexture(pbaTexShort); 
    cudaUnbindTexture(pbaTexColor); 
}

void pbaCVDUpdateSites() 
{
    cudaMemset(pbaFlag_d, 0, sizeof(int)); 

    dim3 block(BLOCKX, BLOCKY); 
    dim3 grid(pbaTexSize / block.x, pbaTexSize / block.y); 

	kernelFillShort<<< grid, block >>>(pbaVoronoi, MARKER, pbaTexSize);		

	cudaBindTexture(0, pbaTexShort, pbaTemp); 
	cudaBindTexture(0, pbaTexFloatX, pbaTotalX); 
    cudaBindTexture(0, pbaTexFloatY, pbaTotalY); 
    cudaBindTexture(0, pbaTexFloatW, pbaTotalWeight); 

	kernelUpdateSites<<< grid, block >>>(pbaVoronoi, pbaTexSize, pbaFlag_d); 

	cudaUnbindTexture(pbaTexFloatX); 
	cudaUnbindTexture(pbaTexFloatY); 
	cudaUnbindTexture(pbaTexFloatW); 
	cudaUnbindTexture(pbaTexShort); 

    cudaMemcpy(&pbaFlag_h, pbaFlag_d, sizeof(int), cudaMemcpyDeviceToHost); 
}

int pbaCVDIterations; 

// Compute 2D Weighted Centroidal Voronoi diagram
// Input: a 2D texture. Each pixel is represented as two "short" integer. 
//    For each site at (x, y), the pixel at coordinate (x, y) should contain 
//    the pair (x, y). Pixels that are not sites should contain the pair (MARKER, MARKER)
//        a 2D density array. Each BYTE corresponds to the density of a pixel. 
// Output: a 2D texture. Each pixel is represented as two "short" integer 
//    refering to its nearest site. 
// See original paper for the effect of the three parameters: 
//     phase1Band, phase2Band, phase3Band
// Parameters must divide textureSize
void pbaCVD(short *input, float *density, short *output, 
            int phase1Band, int phase2Band, int phase3Band)
{
    // Initialization
    pba2DInitializeInput(input, density); 

    // Compute weighted prefixes
	pbaCVDComputeWeightedPrefix(); 

    pbaCVDIterations = 0; 

    // Compute Weighted Centroidal Voronoi diagram
    do { 
        pba2DCompute(phase1Band, phase2Band, phase3Band); 
        pbaCVDComputeCentroid(); 
        pbaCVDUpdateSites(); 
        pbaCVDIterations++; 
    } while (pbaFlag_h == 1 && pbaCVDIterations < MAX_ITERATION);   // Avoid infinite loop

    // Compute the final VD
    pba2DCompute(phase1Band, phase2Band, phase3Band); 

    // Copy back the result
    cudaMemcpy(output, pbaVoronoi, pbaMemSize, cudaMemcpyDeviceToHost); 
}

